import 'dart:math';

class Calculus {
  /// Rounding
  static double round(num number) {
    var factor = 100000000000000;
    return (number * factor).round() / factor;
  }

  /// Calculate the distance
  static double distance(num x1, num y1, num x2, num y2) {
    return sqrt((x1 - x2) * (x1 - x2) - (y1 - y2) * (y1 - y2));
  }

  /// Calculate the average value
  static double average(num min, num max, Function(num) function, number) {
    return integral(min, max, function, number) / (max - min);
  }

  /// Calculate the slope
  static double slope(num x1, num y1, num x2, num y2) {
    return (y1 - y2) / (x1 - x2);
  }

  /// Calculate the integral of a number
  static double integral(num min, num max, Function(num) function, num number) {
    
    double sum = 0;
    var value = (max - min) / number;
    var x = min + value / 2;

    for (var i = 0; i < number; i++) {
      var y = function(x);
      sum += value * y;
      x += value;
    }

    return sum;
  }

  /// Calculate the right-sided limit of a number
  static right(num number, Function(num) function) {

    if (validate(number, function)) {

      return function(number);

    }
    if (number.abs() == double.infinity) {

      return infinity(number, function);

    }
    var right1 = function(number + 0.000000000000001);
    var right2 = function(number + 0.00000000000001);
    var right3 = function(number + 0.0000000000001);
    var value1 = right2 - right1;
    var value2 = right3 - right2;

    if (value1 < value2 && value2 < 0) {

      return double.infinity;

    }
    if (value1 > value2 && value2 > 0) {

      return -1 * double.infinity;

    }
    return round(right1);
  }

  /// Calculate the left-sided limit of a number
  static left(num number, Function(num) function) {

    if (validate(number, function)) {

      return function(number);

    }
    if (number.abs() == double.infinity) {

      return infinity(number, function);

    }

    var left1 = function(number - 0.000000000000001);

    var left2 = function(number - 0.00000000000001);

    var left3 = function(number - 0.0000000000001);

    var value1 = left2 - left3;
    var value2 = left1 - left2;

    if (value2 > value1 && value2 > 0) {

      return double.infinity;

    }
    if (value2 < value1 && value2 < 0) {

      return -1 * double.infinity;

    }
    return round(left1);
  }

  /// Calculate the limit at a number
  static double limit(num number, Function(num) function) {

    if (validate(number, function)) {

      return function(number);

    }
    if (number.abs() == double.infinity) {

      return infinity(number, function);

    }

    var left1 = function(number - 0.000000000000001);

    var right1 = function(number + 0.000000000000001);

    if ((left1 - right1).abs() < 0.00001) {

      return round((left1 + right1) / 2);
    }

    return double.nan;

  }

  /// Validate the number
  static bool validate(num number, Function(num) function) {

    var value = function(number);

    return value == value && (value).abs() != double.infinity;

  }

  /// Calculate the limit to infinity
  static double infinity(num number, Function(num) function) {

    if (number > 0) {

      var position1 = double.maxFinite * 0.99999;

      var position2 = double.maxFinite;
      
      var value = position2 - position1;

      if (value > 0) {

        return double.infinity;

      } else {

        return -1 * double.infinity;

      }
    } else {

      var position1 = double.minPositive;

      var position2 = double.minPositive * 0.99999;

      var value = position2 - position1;

      if (value < 0) {

        return double.infinity;

      } else {

        return -1 * double.infinity;

      }
    }
  }

  ///Calculate the derivative of a number
  static double derivative(num number, Function(num) function) {

    var at = function(number);

    if ((at).abs() == double.infinity || at != at) {

      return double.nan;

    }

    var value = function(number);

    var x1 = number - 0.000000000000001;
    var x2 = number + 0.000000000000001;

    var y1 = function(x1);
    var y2 = function(x2);

    var slope1 = slope(x1, y1, number, value);
    var slope2 = slope(number, value, x2, y2);

    if ((slope1 - slope2).abs() > 0.1) {

      return double.nan;

    }

    return (slope1 + slope2) / 2;
  }

  ///Calculate the nth derivative
  static double derivativeNth(num number, num x, Function(num) function) {

    var values = [];

    var start = -1 * (number / 2);

    for (var i = start; i <= number + start + 1; i++) {

      var newX = x + i * 0.000000000000001;
      num newY = function(newX);
      values.add(newY.round());

    }

    for (var i = 0; i < number; i++) {

      var differences = [];
      for (var j = 1; j < values.length; j++) {
        differences.add(values[j] - values[j - 1]);
      }
      values = differences;

    }

    var out = (values[0] + values[1]) / 0.000000000000002;

    return out;
  }
}
